@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2006, 2007 Canonical
+# Copyright (c) 2006-2009 Canonical
 #
 # Written by Gustavo Niemeyer <gustavo@niemeyer.net>
 #
@@ -18,51 +18,106 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+
+from __future__ import print_function
+
 from datetime import datetime, date, time, timedelta
+import json
+
+from distutils.version import LooseVersion
 
 from storm.databases import dummy
 
+# PostgreSQL support in Storm requires psycopg2 2.3.0 or greater, in order
+# to support the two-phase commit protocol.
+REQUIRED_PSYCOPG2_VERSION = LooseVersion('2.3.0')
+PSYCOPG2_VERSION = None
 try:
     import psycopg2
-    import psycopg2.extensions
-except:
+    PSYCOPG2_VERSION = LooseVersion(psycopg2.__version__.split(' ')[0])
+    if PSYCOPG2_VERSION < REQUIRED_PSYCOPG2_VERSION:
+        psycopg2 = dummy
+    else:
+        import psycopg2.extensions
+except ImportError:
     psycopg2 = dummy
 
 from storm.expr import (
     Undef, Expr, SetExpr, Select, Insert, Alias, And, Eq, FuncExpr, SQLRaw,
-    Sequence, Like, SQLToken, COLUMN, COLUMN_NAME, COLUMN_PREFIX, TABLE,
-    compile, compile_select, compile_insert, compile_set_expr, compile_like,
-    compile_sql_token)
-from storm.variables import Variable, ListVariable, RawStrVariable
-from storm.database import Database, Connection, Result
+    Sequence, Like, SQLToken, BinaryOper, COLUMN, COLUMN_NAME, COLUMN_PREFIX,
+    TABLE, compile, compile_select, compile_insert, compile_set_expr,
+    compile_like, compile_sql_token)
+from storm.variables import (
+    Variable, ListVariable, JSONVariable as BaseJSONVariable)
+from storm.properties import SimpleProperty
+from storm.database import Database, Connection, ConnectionWrapper, Result
 from storm.exceptions import (
-    install_exceptions, DatabaseError, DatabaseModuleError,
-    OperationalError, ProgrammingError, TimeoutError)
+    DatabaseError, DatabaseModuleError, InterfaceError,
+    OperationalError, ProgrammingError, TimeoutError, Error, wrap_exceptions)
 from storm.tracer import TimeoutTracer
 
 
-install_exceptions(psycopg2)
 compile = compile.create_child()
 
 
 class Returning(Expr):
-    """Appends the "RETURNING <primary_columns>" suffix to an INSERT.
+    """Appends the "RETURNING <columns>" suffix to an INSERT or UPDATE.
+
+    @param expr: an L{Insert} or L{Update} expression.
+    @param columns: The columns to return, if C{None} then
+        C{expr.primary_columns} will be used.
 
     This is only supported in PostgreSQL 8.2+.
     """
-
-    def __init__(self, insert):
-        self.insert = insert
+    def __init__(self, expr, columns=None):
+        self.expr = expr
+        self.columns = columns
 
 @compile.when(Returning)
 def compile_returning(compile, expr, state):
     state.push("context", COLUMN)
-    columns = compile(expr.insert.primary_columns, state)
+    columns = expr.columns or expr.expr.primary_columns
+    columns = compile(columns, state)
     state.pop()
     state.push("precedence", 0)
-    insert = compile(expr.insert, state)
+    expr = compile(expr.expr, state)
     state.pop()
-    return "%s RETURNING %s" % (insert, columns)
+    return "%s RETURNING %s" % (expr, columns)
+
+
+class Case(Expr):
+    """A CASE statement.
+
+    @params cases: a list of tuples of (condition, result) or (value, result),
+        if an expression is passed too.
+    @param expression: the expression to compare (if the simple form is used).
+    @param default: an optional default condition if no other case matches.
+
+    """
+    def __init__(self, cases, expression=Undef, default=Undef):
+        self.cases = cases
+        self.expression = expression
+        self.default = default
+
+
+@compile.when(Case)
+def compile_case(compile, expr, state):
+    cases = [
+        "WHEN %s THEN %s" % (
+            compile(condition, state), compile(value, state))
+        for condition, value in expr.cases]
+
+    if expr.expression is not Undef:
+        expression = compile(expr.expression, state) + " "
+    else:
+        expression = ""
+
+    if expr.default is not Undef:
+        default = " ELSE %s" % compile(expr.default, state)
+    else:
+        default = ""
+
+    return "CASE %s%s%s END" % (expression, " ".join(cases), default)
 
 
 class currval(FuncExpr):
@@ -193,24 +248,6 @@ def compile_sql_token_postgres(compile, expr, state):
     return compile_sql_token(compile, expr, state)
 
 
-def compile_str_variable_with_E(compile, variable, state):
-    """Include an E just before the placeholder of string variables.
-
-    PostgreSQL 8.2 will issue a warning without it, and psycopg
-    will use the plain '' rather than E''.  The problem is being
-    tracked at the following URL::
-
-        http://www.initd.org/tracker/psycopg/ticket/202
-
-    """
-    state.parameters.append(variable)
-    if type(variable.get(to_db=True)) is str:
-        return "E?"
-    return "?"
-
-psycopg_needs_E = None
-
-
 class PostgresResult(Result):
 
     def get_insert_identity(self, primary_key, primary_variables):
@@ -225,6 +262,18 @@ class PostgresResult(Result):
         return And(*equals)
 
 
+pg_connection_failure_codes = frozenset([
+    '08006',  # CONNECTION FAILURE
+    '08001',  # SQLCLIENT UNABLE TO ESTABLISH SQLCONNECTION
+    '08004',  # SQLSERVER REJECTED ESTABLISHMENT OF SQLCONNECTION
+    '53300',  # TOO MANY CONNECTIONS
+    '57000',  # OPERATOR INTERVENTION
+    '57P01',  # ADMIN SHUTDOWN
+    '57P02',  # CRASH SHUTDOWN
+    '57P03',  # CANNOT CONNECT NOW
+    ])
+
+
 class PostgresConnection(Connection):
 
     result_factory = PostgresResult
@@ -256,53 +305,65 @@ class PostgresConnection(Connection):
 
         return Connection.execute(self, statement, params, noresult)
 
-    def raw_execute(self, statement, params):
-        """
-        Like L{Connection.raw_execute}, but encode the statement to
-        UTF-8 if it is unicode.
-        """
-        if type(statement) is unicode:
-            # psycopg breaks with unicode statements.
-            statement = statement.encode("UTF-8")
-        return Connection.raw_execute(self, statement, params)
-
     def to_database(self, params):
         """
         Like L{Connection.to_database}, but this converts datetime
-        types to strings, unicode to UTF-8 encoded strings, and
-        strings to L{psycopg2.Binary} instances.
+        types to strings, and bytes to L{psycopg2.Binary} instances.
         """
         for param in params:
             if isinstance(param, Variable):
                 param = param.get(to_db=True)
             if isinstance(param, (datetime, date, time, timedelta)):
                 yield str(param)
-            elif isinstance(param, unicode):
-                yield param.encode("UTF-8")
-            elif isinstance(param, str):
+            elif isinstance(param, bytes):
                 yield psycopg2.Binary(param)
             else:
                 yield param
 
-    def is_disconnection_error(self, exc):
-        if not isinstance(exc, (OperationalError, ProgrammingError)):
-            return False
+    def is_disconnection_error(self, exc, extra_disconnection_errors=()):
+        # Attempt to use pgcode to determine the nature of the error. This is
+        # more reliable than string matching because it is not affected by
+        # locale settings. Fall through if pgcode is not available.
+        if isinstance(exc, Error):
+            pgcode = getattr(exc, "pgcode", None)
+            if pgcode in pg_connection_failure_codes:
+                return True
 
-        # XXX: 2007-09-17 jamesh
-        # I have no idea why I am seeing the last exception message
-        # after upgrading to Gutsy.
-        msg = str(exc)
-        return ("server closed the connection unexpectedly" in msg or
-                "could not connect to server" in msg or
-                "no connection to the server" in msg or
+        disconnection_errors = (
+            DatabaseError, InterfaceError, OperationalError,
+            ProgrammingError, extra_disconnection_errors)
+
+        if isinstance(exc, disconnection_errors):
+            # When the connection is closed by a termination of pgbouncer, a
+            # DatabaseError or subclass with no message (depending on
+            # psycopg2 version) is raised. If the raw connection is closed
+            # we assume it's actually a disconnection.
+            if isinstance(exc, DatabaseError):
+                if self._raw_connection.closed:
+                    return True
+            msg = str(exc)
+            return (
+                "SSL SYSCALL error" in msg or
+                "EOF detected" in msg or
+                "connection already closed" in msg or
                 "connection not open" in msg or
-                "losed the connection unexpectedly" in msg)
+                "could not connect to server" in msg or
+                "could not receive data from server" in msg or
+                "could not send data to server" in msg or
+                "losed the connection unexpectedly" in msg or
+                "no connection to the server" in msg or
+                "server closed the connection unexpectedly" in msg or
+                "terminating connection due to administrator" in msg)
+
+        return False
 
 
 class Postgres(Database):
 
     connection_factory = PostgresConnection
 
+    _exception_module = psycopg2
+
     # An integer representing the server version.  If the server does
     # not support the server_version_num variable, this will be set to
     # 0.  In practice, this means the variable will be 0 or greater
@@ -310,15 +371,22 @@ class Postgres(Database):
     _version = None
 
     def __init__(self, uri):
+        super(Postgres, self).__init__(uri)
         if psycopg2 is dummy:
-            raise DatabaseModuleError("'psycopg2' module not found")
+            raise DatabaseModuleError(
+                "'psycopg2' >= %s not found. Found %s."
+                % (REQUIRED_PSYCOPG2_VERSION, PSYCOPG2_VERSION))
         self._dsn = make_dsn(uri)
-        isolation = uri.options.get("isolation", "serializable")
+        isolation = uri.options.get("isolation", "repeatable-read")
         isolation_mapping = {
             "autocommit": psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT,
             "serializable": psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE,
-            "read-committed": psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED,
-        }
+            "read-committed":
+                psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED,
+            "repeatable-read":
+                psycopg2.extensions.ISOLATION_LEVEL_REPEATABLE_READ,
+            "read-uncommitted":
+                psycopg2.extensions.ISOLATION_LEVEL_READ_UNCOMMITTED}
         try:
             self._isolation = isolation_mapping[isolation]
         except KeyError:
@@ -327,45 +395,48 @@ class Postgres(Database):
                 "'autocommit', 'serializable', 'read-committed'" %
                 (isolation,))
 
-    def raw_connect(self):
-        raw_connection = psycopg2.connect(self._dsn)
+    _psycopg_error_attributes = ["pgerror", "pgcode", "cursor"]
+    # Added in psycopg2 2.5.
+    if hasattr(psycopg2.Error, "diag"):
+        _psycopg_error_attributes.append("diag")
+    _psycopg_error_attributes = tuple(_psycopg_error_attributes)
+
+    def _make_combined_exception_type(self, wrapper_type, dbapi_type):
+        combined_type = super(Postgres, self)._make_combined_exception_type(
+            wrapper_type, dbapi_type)
+        for name in self._psycopg_error_attributes:
+            setattr(combined_type, name, lambda err: getattr(err, "_" + name))
+        return combined_type
+
+    def _wrap_exception(self, wrapper_type, exception):
+        wrapped = super(Postgres, self)._wrap_exception(
+            wrapper_type, exception)
+        for name in self._psycopg_error_attributes:
+            setattr(wrapped, "_" + name, getattr(exception, name))
+        return wrapped
+
+    def _raw_connect(self):
+        raw_connection = ConnectionWrapper(psycopg2.connect(self._dsn), self)
 
         if self._version is None:
             cursor = raw_connection.cursor()
 
             try:
                 cursor.execute("SHOW server_version_num")
-            except psycopg2.ProgrammingError:
+            except ProgrammingError:
                 self._version = 0
-                raw_connection.rollback()
             else:
                 self._version = int(cursor.fetchone()[0])
-
-            # This will conditionally change the compilation of binary
-            # variables (RawStrVariable) to preceed the placeholder with an
-            # 'E', if psycopg isn't doing it by itself.
-            #
-            # The "failing" code path isn't unittested because that
-            # would depend on a working psycopg version, or in an older
-            # postgresql version.  Both branches were manually tested
-            # for correctness at some point.
-            global psycopg_needs_E
-            try:
-                # If psycopg behaves correctly, this will break trying
-                # to run a query with EE''.
-                cursor.execute("SELECT E%s", (psycopg2.Binary(""),))
-            except psycopg2.ProgrammingError:
-                psycopg_needs_E = False
-            else:
-                psycopg_needs_E = True
-                compile.when(RawStrVariable)(compile_str_variable_with_E)
-
             raw_connection.rollback()
 
         raw_connection.set_client_encoding("UTF8")
         raw_connection.set_isolation_level(self._isolation)
         return raw_connection
 
+    def raw_connect(self):
+        with wrap_exceptions(self):
+            return self._raw_connect()
+
 
 create_from_uri = Postgres
 
@@ -401,4 +472,40 @@ class PostgresTimeoutTracer(TimeoutTracer):
         # psycopg2.extensions.QueryCanceledError in the future.
         if (isinstance(error, DatabaseError) and
             "statement timeout" in str(error)):
-            raise TimeoutError(statement, params)
+            raise TimeoutError(
+                statement, params, "SQL server cancelled statement")
+
+
+# Postgres-specific operators
+
+class JSONElement(BinaryOper):
+    """Return an element of a JSON value (by index or field name)."""
+
+    oper = "->"
+
+
+class JSONTextElement(BinaryOper):
+    """Return an element of a JSON value (by index or field name) as text."""
+
+    oper = "->>"
+
+
+# Postgres-specific properties and variables
+
+class JSONVariable(BaseJSONVariable):
+
+    __slots__ = ()
+
+    def _loads(self, value):
+        if isinstance(value, str):
+            # psycopg versions < 2.5 don't automatically convert JSON columns
+            # to python objects, they return a string.
+            #
+            # Note that on newer versions, if the object contained is an actual
+            # string, it's returned as unicode, so the check is still valid.
+            return json.loads(value)
+        return value
+
+
+class JSON(SimpleProperty):
+    variable_class = JSONVariable
