@@ -18,10 +18,14 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import print_function
+
 from datetime import datetime, date, time, timedelta
 from time import sleep, time as now
 import sys
 
+import six
+
 from storm.databases import dummy
 
 try:
@@ -32,23 +36,27 @@ except ImportError:
     except ImportError:
         sqlite = dummy
 
-from storm.variables import Variable, RawStrVariable
-from storm.database import Database, Connection, Result
-from storm.exceptions import install_exceptions, DatabaseModuleError
+from storm.variables import Variable, BytesVariable
+from storm.database import Database, Connection, ConnectionWrapper, Result
+from storm.exceptions import (
+    DatabaseModuleError, OperationalError, wrap_exceptions)
 from storm.expr import (
     Insert, Select, SELECT, Undef, SQLRaw, Union, Except, Intersect,
     compile, compile_insert, compile_select)
 
 
-install_exceptions(sqlite)
-
-
 compile = compile.create_child()
 
 @compile.when(Select)
 def compile_select_sqlite(compile, select, state):
     if select.offset is not Undef and select.limit is Undef:
-        select.limit = sys.maxint
+        if sys.maxsize > 2**32:
+            # On 64-bit platforms sqlite doesn't like maxsize as LIMIT. See
+            # also
+            # https://lists.ubuntu.com/archives/storm/2013-June/001492.html
+            select.limit = sys.maxsize - 1
+        else:
+            select.limit = sys.maxsize
     statement = compile_select(compile, select, state)
     if state.context is SELECT:
         # SQLite breaks with (SELECT ...) UNION (SELECT ...), so we
@@ -77,21 +85,23 @@ class SQLiteResult(Result):
 
     @staticmethod
     def set_variable(variable, value):
-        if isinstance(variable, RawStrVariable):
+        if (isinstance(variable, BytesVariable) and
+                isinstance(value, six.text_type)):
             # pysqlite2 may return unicode.
-            value = str(value)
+            value = value.encode("UTF-8")
         variable.set(value, from_db=True)
 
     @staticmethod
     def from_database(row):
-        """Convert MySQL-specific datatypes to "normal" Python types.
+        """Convert SQLite-specific datatypes to "normal" Python types.
 
-        If there are anny C{buffer} instances in the row, convert them
-        to strings.
+        On Python 2, if there are any C{buffer} instances in the row,
+        convert them to bytes.  On Python 3, BLOB types are converted to
+        bytes, which is already what we want.
         """
         for value in row:
-            if isinstance(value, buffer):
-                yield str(value)
+            if six.PY2 and isinstance(value, buffer):
+                yield bytes(value)
             else:
                 yield value
 
@@ -106,7 +116,7 @@ class SQLiteConnection(Connection):
     def to_database(params):
         """
         Like L{Connection.to_database}, but this also converts
-        instances of L{datetime} types to strings, and strings
+        instances of L{datetime} types to strings, and (on Python 2) bytes
         instances to C{buffer} instances.
         """
         for param in params:
@@ -114,12 +124,13 @@ class SQLiteConnection(Connection):
                 param = param.get(to_db=True)
             if isinstance(param, (datetime, date, time, timedelta)):
                 yield str(param)
-            elif isinstance(param, str):
+            elif six.PY2 and isinstance(param, bytes):
                 yield buffer(param)
             else:
                 yield param
 
     def commit(self):
+        self._ensure_connected()
         # See story at the end to understand why we do COMMIT manually.
         if self._in_transaction:
             self.raw_execute("COMMIT", _end=True)
@@ -151,7 +162,7 @@ class SQLiteConnection(Connection):
         while True:
             try:
                 return Connection.raw_execute(self, statement, params)
-            except sqlite.OperationalError, e:
+            except OperationalError as e:
                 if str(e) != "database is locked":
                     raise
                 elif now() - started < self._database._timeout:
@@ -171,17 +182,41 @@ class SQLite(Database):
 
     connection_factory = SQLiteConnection
 
+    _exception_module = sqlite
+
     def __init__(self, uri):
+        super(SQLite, self).__init__(uri)
         if sqlite is dummy:
             raise DatabaseModuleError("'pysqlite2' module not found")
         self._filename = uri.database or ":memory:"
         self._timeout = float(uri.options.get("timeout", 5))
+        self._synchronous = uri.options.get("synchronous")
+        self._journal_mode = uri.options.get("journal_mode")
+        self._foreign_keys = uri.options.get("foreign_keys")
+
+    def _raw_connect(self):
+        # See the story at the end to understand why we set isolation_level.
+        raw_connection = ConnectionWrapper(
+            sqlite.connect(
+                self._filename, timeout=self._timeout, isolation_level=None),
+            self)
+        if self._synchronous is not None:
+            raw_connection.execute("PRAGMA synchronous = %s" %
+                                   (self._synchronous,))
+
+        if self._journal_mode is not None:
+            raw_connection.execute("PRAGMA journal_mode = %s" %
+                                   (self._journal_mode,))
+
+        if self._foreign_keys is not None:
+            raw_connection.execute("PRAGMA foreign_keys = %s" %
+                                   (self._foreign_keys,))
+
+        return raw_connection
 
     def raw_connect(self):
-        # See the story at the end to understand why we set isolation_level.
-        raw_connection = sqlite.connect(self._filename, timeout=self._timeout,
-                                        isolation_level=None)
-        return raw_connection
+        with wrap_exceptions(self):
+            return self._raw_connect()
 
 
 create_from_uri = SQLite
@@ -222,3 +257,18 @@ create_from_uri = SQLite
 #     http://www.sqlite.org/lockingv3.html
 #     http://docs.python.org/lib/sqlite3-Controlling-Transactions.html
 #
+
+# --------------------------------------------------------------------
+# Reserved words, SQLite specific
+
+# The list of reserved words here are SQLite specific.  SQL92 reserved words
+# are registered in storm.expr, near the "Reserved words, from SQL1992"
+# comment.  The reserved words here were taken from:
+#
+# http://www.sqlite.org/lang_keywords.html
+compile.add_reserved_words("""
+    abort after analyze attach autoincrement before conflict database detach
+    each exclusive explain fail glob if ignore index indexed instead isnull
+    limit notnull offset plan pragma query raise regexp reindex release
+    rename replace row savepoint temp trigger vacuum virtual
+    """.split())
