@@ -18,15 +18,19 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
+from __future__ import print_function
+
 from decimal import Decimal
 from datetime import datetime, date, time, timedelta
 from weakref import WeakKeyDictionary
 from copy import copy
 import re
 
+import six
+
 from storm.exceptions import CompileError, NoTableError, ExprError
 from storm.variables import (
-    Variable, RawStrVariable, UnicodeVariable, LazyValue,
+    Variable, BytesVariable, UnicodeVariable, LazyValue,
     DateTimeVariable, DateVariable, TimeVariable, TimeDeltaVariable,
     BoolVariable, IntVariable, FloatVariable, DecimalVariable)
 from storm import Undef, has_cextensions
@@ -79,10 +83,10 @@ class Compile(object):
 
         Use this as:
 
-        @compile.when(TypeA, TypeB)
-        def compile_type_a_or_b(compile, expr, state):
-            ...
-            return "THE COMPILED SQL STATEMENT"
+            >>> @compile.when(TypeA, TypeB)
+            >>> def compile_type_a_or_b(compile, expr, state):
+            >>>     ...
+            >>>     return "THE COMPILED SQL STATEMENT"
         """
         return _when(self, types)
 
@@ -153,21 +157,23 @@ class Compile(object):
             created internally (and thus can't be accessed).
         @param join: The string token to use to put between
             subexpressions. Defaults to ", ".
-        @param raw: If true, any string or unicode expression or
-            subexpression will not be further compiled.
-        @param token: If true, any string or unicode expression will
-            be considered as a SQLToken, and quoted properly.
+        @param raw: If true, any string (str or unicode on Python 2, str on
+            Python 3) expression or subexpression will not be further
+            compiled.
+        @param token: If true, any string (str or unicode on Python 2, str
+            on Python 3) expression will be considered as a SQLToken, and
+            quoted properly.
         """
         # FASTPATH This method is part of the fast path.  Be careful when
         #          changing it (try to profile any changes).
 
         expr_type = type(expr)
+        string_types = (str,) if six.PY3 else (str, unicode)
 
-        if (expr_type is SQLRaw or
-            raw and (expr_type is str or expr_type is unicode)):
+        if expr_type is SQLRaw or (raw and expr_type in string_types):
             return expr
 
-        if token and (expr_type is str or expr_type is unicode):
+        if token and expr_type in string_types:
             expr = SQLToken(expr)
 
         if state is None:
@@ -178,15 +184,14 @@ class Compile(object):
             compiled = []
             for subexpr in expr:
                 subexpr_type = type(subexpr)
-                if subexpr_type is SQLRaw or raw and (subexpr_type is str or
-                                                      subexpr_type is unicode):
+                if (subexpr_type is SQLRaw or
+                        (raw and subexpr_type in string_types)):
                     statement = subexpr
                 elif subexpr_type is tuple or subexpr_type is list:
                     state.precedence = outer_precedence
                     statement = self(subexpr, state, join, raw, token)
                 else:
-                    if token and (subexpr_type is unicode or
-                                  subexpr_type is str):
+                    if token and subexpr_type in string_types:
                         subexpr = SQLToken(subexpr)
                     statement = self._compile_single(subexpr, state,
                                                      outer_precedence)
@@ -216,7 +221,7 @@ class CompilePython(Compile):
                 "    return match" %
                 (",".join("_%d" % i for i in range(len(state.parameters))),
                  source))
-        exec code in namespace
+        exec(code, namespace)
         return namespace['closure'](state.parameters, bool)
 
 
@@ -303,17 +308,17 @@ SELECT = Context("SELECT")
 # --------------------------------------------------------------------
 # Builtin type support
 
-@compile.when(str)
-def compile_str(compile, expr, state):
-    state.parameters.append(RawStrVariable(expr))
+@compile.when(bytes)
+def compile_bytes(compile, expr, state):
+    state.parameters.append(BytesVariable(expr))
     return "?"
 
-@compile.when(unicode)
-def compile_unicode(compile, expr, state):
+@compile.when(six.text_type)
+def compile_text(compile, expr, state):
     state.parameters.append(UnicodeVariable(expr))
     return "?"
 
-@compile.when(int, long)
+@compile.when(*six.integer_types)
 def compile_int(compile, expr, state):
     state.parameters.append(IntVariable(expr))
     return "?"
@@ -358,7 +363,8 @@ def compile_none(compile, expr, state):
     return "NULL"
 
 
-@compile_python.when(str, unicode, int, long, float, type(None))
+@compile_python.when(
+    bytes, six.text_type, float, type(None), *six.integer_types)
 def compile_python_builtin(compile, expr, state):
     return repr(expr)
 
@@ -395,8 +401,18 @@ def compile_python_unsupported(compile, expr, state):
     raise CompileError("Can't compile python expressions with %r" % type(expr))
 
 
+# A translation table that can escape a unicode string for use in a
+# Like() expression that uses "!" as the escape character.
+like_escape = {
+    ord(u"!"): u"!!",
+    ord(u"_"): u"!_",
+    ord(u"%"): u"!%"
+    }
+
+
 class Comparable(object):
     __slots__ = ()
+    __hash__ = object.__hash__
 
     def __eq__(self, other):
         if other is not None and not isinstance(other, (Expr, Variable)):
@@ -468,6 +484,9 @@ class Comparable(object):
             other = getattr(self, "variable_factory", Variable)(value=other)
         return Div(self, other)
 
+    __floordiv__ = __div__
+    __truediv__ = __div__
+
     def __mod__(self, other):
         if not isinstance(other, (Expr, Variable)):
             other = getattr(self, "variable_factory", Variable)(value=other)
@@ -498,6 +517,24 @@ class Comparable(object):
     def upper(self):
         return Upper(self)
 
+    def startswith(self, prefix):
+        if not isinstance(prefix, six.text_type):
+            raise ExprError("Expected text argument, got %r" % type(prefix))
+        pattern = prefix.translate(like_escape) + u"%"
+        return Like(self, pattern, u"!")
+
+    def endswith(self, suffix):
+        if not isinstance(suffix, six.text_type):
+            raise ExprError("Expected text argument, got %r" % type(suffix))
+        pattern = u"%" + suffix.translate(like_escape)
+        return Like(self, pattern, u"!")
+
+    def contains_string(self, substring):
+        if not isinstance(substring, six.text_type):
+            raise ExprError("Expected text argument, got %r" % type(substring))
+        pattern = u"%" + substring.translate(like_escape) + u"%"
+        return Like(self, pattern, u"!")
+
 
 class ComparableExpr(Expr, Comparable):
     __slots__ = ()
@@ -629,10 +666,13 @@ class Select(Expr):
 @compile.when(Select)
 def compile_select(compile, select, state):
     tokens = ["SELECT "]
-    if select.distinct:
-        tokens.append("DISTINCT ")
     state.push("auto_tables", [])
     state.push("context", COLUMN)
+    if select.distinct:
+        tokens.append("DISTINCT ")
+        if isinstance(select.distinct, (tuple, list)):
+            tokens.append(
+                "ON (%s) " % compile(select.distinct, state, raw=True))
     tokens.append(compile(select.columns, state))
     tables_pos = len(tokens)
     parameters_pos = len(state.parameters)
@@ -670,7 +710,8 @@ def compile_select(compile, select, state):
 class Insert(Expr):
     """Expression representing an insert statement.
 
-    @ivar map: Dictionary mapping columns to values.
+    @ivar map: Dictionary mapping columns to values, or a sequence of columns
+        for a bulk insert.
     @ivar table: Table where the row should be inserted.
     @ivar default_table: Table to use if no table is explicitly provided, and
         no tables may be inferred from provided columns.
@@ -680,17 +721,21 @@ class Insert(Expr):
     @ivar primary_variables: Tuple of variables with values for the primary
         key of the table where the row will be inserted.  This is a hint used
         by backends to process the insertion of rows.
+    @ivar values: Expression or sequence of tuples of values for bulk
+        insertion.
     """
     __slots__ = ("map", "table", "default_table", "primary_columns",
-                 "primary_variables")
+                 "primary_variables", "values")
 
     def __init__(self, map, table=Undef, default_table=Undef,
-                 primary_columns=Undef, primary_variables=Undef):
+                 primary_columns=Undef, primary_variables=Undef,
+                 values=Undef):
         self.map = map
         self.table = table
         self.default_table = default_table
         self.primary_columns = primary_columns
         self.primary_variables = primary_variables
+        self.values = values
 
 @compile.when(Insert)
 def compile_insert(compile, insert, state):
@@ -699,20 +744,30 @@ def compile_insert(compile, insert, state):
     state.context = TABLE
     table = build_tables(compile, insert.table, insert.default_table, state)
     state.context = EXPR
-    values = compile(tuple(insert.map.itervalues()), state)
+    values = insert.values
+    if values is Undef:
+        values = [tuple(six.itervalues(insert.map))]
+    if isinstance(values, Expr):
+        compiled_values = compile(values, state)
+    else:
+        compiled_values = (
+            "VALUES (%s)" %
+            "), (".join(compile(value, state) for value in values))
     state.pop()
-    return "".join(["INSERT INTO ", table, " (", columns,
-                    ") VALUES (", values, ")"])
+    return "".join(
+        ["INSERT INTO ", table, " (", columns, ") ", compiled_values])
 
 
 class Update(Expr):
-    __slots__ = ("map", "where", "table", "default_table")
+    __slots__ = ("map", "where", "table", "default_table", "primary_columns")
 
-    def __init__(self, map, where=Undef, table=Undef, default_table=Undef):
+    def __init__(self, map, where=Undef, table=Undef, default_table=Undef,
+                 primary_columns=Undef):
         self.map = map
         self.where = where
         self.table = table
         self.default_table = default_table
+        self.primary_columns = primary_columns
 
 @compile.when(Update)
 def compile_update(compile, update, state):
@@ -770,7 +825,8 @@ class Column(ComparableExpr):
     @ivar variable_factory: Factory producing C{Variable} instances typed
         according to this column.
     """
-    __slots__ = ("name", "table", "primary", "variable_factory")
+    __slots__ = ("name", "table", "primary", "variable_factory",
+                 "compile_cache", "compile_id")
 
     def __init__(self, name=Undef, table=Undef, primary=False,
                  variable_factory=None):
@@ -778,6 +834,8 @@ class Column(ComparableExpr):
         self.table = table
         self.primary = int(primary)
         self.variable_factory = variable_factory or Variable
+        self.compile_cache = None
+        self.compile_id = None
 
 @compile.when(Column)
 def compile_column(compile, column, state):
@@ -789,11 +847,17 @@ def compile_column(compile, column, state):
             alias = state.aliases.get(column)
             if alias is not None:
                 return compile(alias.name, state, token=True)
-        return compile(column.name, state, token=True)
+        if column.compile_id != id(compile):
+            column.compile_cache = compile(column.name, state, token=True)
+            column.compile_id = id(compile)
+        return column.compile_cache
     state.push("context", COLUMN_PREFIX)
     table = compile(column.table, state, token=True)
     state.pop()
-    return "%s.%s" % (table, compile(column.name, state, token=True))
+    if column.compile_id != id(compile):
+        column.compile_cache = compile(column.name, state, token=True)
+        column.compile_id = id(compile)
+    return "%s.%s" % (table, column.compile_cache)
 
 @compile_python.when(Column)
 def compile_python_column(compile, column, state):
@@ -840,14 +904,20 @@ class FromExpr(Expr):
 
 
 class Table(FromExpr):
-    __slots__ = ("name",)
+    __slots__ = ("name", "compile_cache", "compile_id")
 
     def __init__(self, name):
         self.name = name
+        self.compile_cache = None
+        self.compile_id = None
+
 
 @compile.when(Table)
 def compile_table(compile, table, state):
-    return compile(table.name, state, token=True)
+    if table.compile_id != id(compile):
+        table.compile_cache = compile(table.name, state, token=True)
+        table.compile_id = id(compile)
+    return table.compile_cache
 
 
 class JoinExpr(FromExpr):
@@ -922,6 +992,21 @@ class NaturalRightJoin(JoinExpr):
     oper = "NATURAL RIGHT JOIN"
 
 
+# --------------------------------------------------------------------
+# Distinct expressions
+
+class Distinct(Expr):
+    """Add the 'DISTINCT' prefix to an expression."""
+    __slots__ = ("expr")
+
+    def __init__(self, expr):
+        self.expr = expr
+
+@compile.when(Distinct)
+def compile_distinct(compile, distinct, state):
+    return "DISTINCT %s" % compile(distinct.expr, state)
+
+
 # --------------------------------------------------------------------
 # Operators
 
@@ -1006,11 +1091,11 @@ class Le(BinaryOper):
 
 class RShift(BinaryOper):
     __slots__ = ()
-    oper = ">>"
+    oper = " >> "
 
 class LShift(BinaryOper):
     __slots__ = ()
-    oper = "<<"
+    oper = " << "
 
 
 class Like(BinaryOper):
@@ -1054,23 +1139,23 @@ def compile_in(compile, expr, state):
 
 class Add(CompoundOper):
     __slots__ = ()
-    oper = "+"
+    oper = " + "
 
 class Sub(NonAssocBinaryOper):
     __slots__ = ()
-    oper = "-"
+    oper = " - "
 
 class Mul(CompoundOper):
     __slots__ = ()
-    oper = "*"
+    oper = " * "
 
 class Div(NonAssocBinaryOper):
     __slots__ = ()
-    oper = "/"
+    oper = " / "
 
 class Mod(NonAssocBinaryOper):
     __slots__ = ()
-    oper = "%"
+    oper = " % "
 
 
 class And(CompoundOper):
@@ -1099,6 +1184,16 @@ class SetExpr(Expr):
         self.order_by = kwargs.get("order_by", Undef)
         self.limit = kwargs.get("limit", Undef)
         self.offset = kwargs.get("offset", Undef)
+        # If the first expression is of a compatible type, directly
+        # include its sub expressions.
+        if len(self.exprs) > 0:
+            first = self.exprs[0]
+            if (isinstance(first, self.__class__) and
+                first.all == self.all and
+                first.limit is Undef and
+                first.offset is Undef):
+                self.exprs = first.exprs + self.exprs[1:]
+
 
 @compile.when(SetExpr)
 def compile_set_expr(compile, expr, state):
@@ -1249,6 +1344,31 @@ class Coalesce(NamedFunc):
     name = "COALESCE"
 
 
+class Row(NamedFunc):
+    __slots__ = ()
+    name = "ROW"
+
+
+class Cast(FuncExpr):
+    """A representation of C{CAST} clauses. e.g., C{CAST(bar AS TEXT)}."""
+    __slots__ = ("column", "type")
+    name = "CAST"
+
+    def __init__(self, column, type):
+        """Create a cast of C{column} as C{type}."""
+        self.column = column
+        self.type = type
+
+
+@compile.when(Cast)
+def compile_cast(compile, cast, state):
+    """Compile L{Cast} expressions."""
+    state.push("context", EXPR)
+    column = compile(cast.column, state)
+    state.pop()
+    return "CAST(%s AS %s)" % (column, cast.type)
+
+
 # --------------------------------------------------------------------
 # Prefix and suffix expressions
 
@@ -1304,7 +1424,7 @@ class Desc(SuffixExpr):
 # --------------------------------------------------------------------
 # Plain SQL expressions.
 
-class SQLRaw(str):
+class SQLRaw(six.text_type):
     """Subtype to mark a string as something that shouldn't be compiled.
 
     This is handled internally by the compiler.
@@ -1312,7 +1432,7 @@ class SQLRaw(str):
     __slots__ = ()
 
 
-class SQLToken(str):
+class SQLToken(six.text_type):
     """Marker for strings that should be considered as a single SQL token.
 
     These strings will be quoted, when needed.
